@@ -1,31 +1,31 @@
 package Crypt::Credentials;
-$Crypt::Credentials::VERSION = '0.002';
+$Crypt::Credentials::VERSION = '0.003';
 use strict;
 use warnings;
 
 use Carp 'croak';
 use Crypt::AuthEnc::GCM qw/gcm_encrypt_authenticate gcm_decrypt_verify/;
 use Crypt::URandom 0.37 'urandom_ub';
+use File::Basename 'dirname';
 use File::Path 'make_path';
 use File::Slurper qw/read_binary write_binary/;
-use File::Spec::Functions qw/catdir catfile curdir rel2abs/;
+use File::Spec::Functions qw/catdir catfile curdir updir abs2rel rel2abs/;
 use YAML::PP;
 
 sub new {
 	my ($class, %args) = @_;
 
 	my $dir = rel2abs($args{dir} // catdir(curdir, 'credentials'));
-	make_path($dir);
 
 	my $check_file = catfile($dir, 'check.enc');
 
 	my $real_key;
 
-	if (-e $check_file) {
+	if (-f $check_file) {
 		for my $key (@{ $args{keys} }) {
 			my $length = length $key;
 			croak "Invalid key size($length)" if $length != 16 && $length != 24 && $length != 32;
-			if (eval { $class->_get($key, $check_file) } // '' eq 'OK') {
+			if (eval { $class->_get($check_file, $key) } // '' eq 'OK') {
 				$real_key = $key;
 				last;
 			}
@@ -34,8 +34,10 @@ sub new {
 		($real_key) = @{ $args{keys} };
 		my $length = length $real_key;
 		croak "Invalid key size($length)" if $length != 16 && $length != 24 && $length != 32;
+		make_path($dir);
 		$class->_put($check_file, $real_key, 'OK');
 	}
+	croak 'No working key found' unless defined $real_key;
 
 	return bless {
 		key => $real_key,
@@ -57,6 +59,8 @@ sub _put {
 sub put {
 	my ($self, $name, $plaintext) = @_;
 	my $filename = catfile($self->{dir}, "$name.yml.enc");
+	my $dirname = dirname($filename);
+	make_path($dirname);
 	$self->_put($filename, $self->{key}, $plaintext);
 	return;
 }
@@ -68,7 +72,7 @@ sub put_yaml {
 }
 
 sub _get {
-	my ($self, $key, $filename) = @_;
+	my ($self, $filename, $key) = @_;
 	my $raw = read_binary($filename);
 	my ($iv, $tag, $ciphertext) = unpack $format, $raw;
 	my $plaintext = gcm_decrypt_verify('AES', $key, $iv, '', $ciphertext, $tag);
@@ -79,8 +83,8 @@ sub _get {
 sub get {
 	my ($self, $name) = @_;
 	my $filename = catfile($self->{dir}, "$name.yml.enc");
-	croak "No such credentials '$name'" if not -e $filename;
-	return $self->_get($self->{key}, $filename);
+	croak "No such credentials '$name'" if not -f $filename;
+	return $self->_get($filename, $self->{key});
 }
 
 sub get_yaml {
@@ -92,8 +96,25 @@ sub get_yaml {
 sub has {
 	my ($self, $name) = @_;
 
-	my $filename = catfile($self->{dir}, "$name.yml.enc");
-	return -e $filename;
+	return -f catfile($self->{dir}, "$name.yml.enc");
+}
+
+sub _recode_dir {
+	my ($self, $dir, $new_key) = @_;
+
+	opendir my $dh, $dir or croak "Could not open dir: $!";
+	while (my $file = readdir $dh) {
+		next if $file eq curdir || $file eq updir;
+		my $filename = catfile($dir, $file);
+
+		if (-d $filename) {
+			$self->_recode_dir($filename, $new_key);
+		} elsif (-f $filename) {
+			next unless $file =~ /\.yml\.enc$/;
+			my $plaintext = $self->_get($filename, $self->{key});
+			$self->_put($filename, $new_key, $plaintext);
+		}
+	}
 }
 
 sub recode {
@@ -102,14 +123,7 @@ sub recode {
 	my $key_length = length $new_key;
 	croak "Invalid key size($key_length)" if $key_length != 16 && $key_length != 24 && $key_length != 32;
 
-	opendir my $dh, $self->{dir} or croak "Could not open dir: $!";
-	while (my $file = readdir $dh) {
-		next unless $file =~ /\.yml\.enc$/;
-		my $filename = catfile($self->{dir}, $file);
-
-		my $plaintext = $self->_get($self->{key}, $filename);
-		$self->_put($filename, $new_key, $plaintext);
-	}
+	$self->_recode_dir($self->{dir}, $new_key);
 
 	my $check_file = catfile($self->{dir}, 'check.enc');
 	$self->_put($check_file, $new_key, 'OK');
@@ -124,11 +138,29 @@ sub remove {
 	return unlink($filename);
 }
 
+sub _list_dir {
+	my ($self, $base, $dir) = @_;
+	opendir my $dh, $dir or croak "No such dir $dir: $!";
+	my @files;
+	while (my $file = readdir $dh) {
+		next if $file eq curdir || $file eq updir;
+		my $filename = catfile($dir, $file);
+
+		if (-d $filename) {
+			push @files, $self->_list_dir($base, $filename);
+		} elsif (-f $filename and $filename =~ s/\.yml\.enc$//) {
+			push @files, abs2rel($filename, $base);
+		}
+	}
+	return @files;
+}
+
 sub list {
-	my ($self) = @_;
-	opendir my $dh, $self->{dir} or croak "No such dir $self->{dir}";
-	my @files = readdir $dh;
-	return grep s/\.yml\.enc$//, @files;
+	my ($self, $base) = @_;
+	my $dir = $base ? catdir($self->{dir}, $base) : $self->{dir};
+	return if not -d $dir;
+
+	return $self->_list_dir($self->{dir}, $dir);
 }
 
 1;
@@ -147,7 +179,7 @@ Crypt::Credentials - Manage credential files
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
@@ -166,9 +198,9 @@ This module implements a credentials store. Essentially it allows you to expand
 
 =head2 new
 
- $self->new(key => $key, dir => $dir)
+ $self->new(keys => \@keys, dir => $dir)
 
-This creates a new C<Crypt::Credentials> object. It takes two named arguments: C<$key> (mandatory) is the cryptographic key used to encrypt the credentials, it must be either 16, 24, or 32 bytes long. C<$dir> is optional for the directory in which the credentials are stored, it defaults to F<./credentials>.
+This creates a new C<Crypt::Credentials> object. It takes two named arguments: C<@keys> (mandatory) are the cryptographic keys used to encrypt the credentials, they must be either 16, 24, or 32 bytes long. If multiple keys are given they're tried until the right one is found, this facilitates key rotation. C<$dir> is optional for the directory in which the credentials are stored, it defaults to F<./credentials>.
 
 =head2 get
 
